/*
Version 1.1, 6/1/1999

This program is intended to create .wav files from CPC files
stored on the PC's hard disk, in order to send them to the "actual" CPC.

Then it launches a .wav player ...

DISCLAIMER : This program comes with no warranty. The author can not
be held responsible for any damage caused directly or indirectly by
the use of this program .
That is, I'm not responsible if it sets your CPU on fire, blows up your
screen or trashes your hard disk.
Well, I don't think it can do anything harmful to your system, but you
never know :).

Written by Pierre Thevenet 27/12/1997

Send comments to p-thev98@bat710.univ-lyon1.fr
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>



/* Widthes of 0 & 1 bits                                              */
/* I previously used that to control transfer speed, but modifying    */
/* sample rate is far more efficient                                  */

#define W0 1
#define W1 2

/* low & high levels in the .wav file                               */
/* LOW must be under 128 and HIGH above this value                  */
/* The nearer 128, the less sound output is loud                    */
/* Typically LOW and HIGH are centered around 128 but I don't think */
/* this is requested.                                               */
/* Note : of course, output .wav file is 8bit.                      */

#define LOW 64
#define HIGH 192

#define kCRCpoly  4129  


/* Functions prototypes */
/* Self-explanatory, ain't them :) */

void putbit(FILE *f, unsigned char b, unsigned long int *filesize);
void putbyte(FILE *f, unsigned char b, unsigned long int *filesize);
void putblock(FILE *f, unsigned char *block, unsigned int size, unsigned long int *filesize);
void putWavHeader(FILE *f);
void writesize(FILE *f, unsigned long int size);
void putCRC(FILE *f, unsigned int CRC, unsigned long int *filesize);
void writename(unsigned char *tape_hdr, unsigned char *disc_hdr);
void putsilence(FILE *f, int length, unsigned long int *filesize);
void usage(char *); 
/* CRC functions were copied from Pierre Guerrier's AIFF Decoder */

unsigned int calcCRC(unsigned char* buffer);
unsigned int CRCupdate(unsigned int CRC, unsigned char new);

/* Global variables */
int noext = 0;    /* Should we create file extensions ?        */
int verbose = 0;  /*           be verbose about what is done ? */
int play = 1;     /*           launch a .wav player ?          */

unsigned int rate = 11025;       /* Sampling rate of the output file (Hz) */
                               /* Higher values can give better transfer speed */
                               /* Too high values will give read errors */
unsigned int silence = 0;    /* Silence duration between blocks (in samples) */
unsigned int blank = 0;          /* Blank between a header and its block (idem)  */
/* Writes a single bit to the file */
void putbit(FILE *f, unsigned char b, unsigned long int *filesize)
{
int i,l;
if (b==0)
        l=W0;
else
        l=W1;
*filesize +=l*2; /* Need the file size for the .wav header */
for (i=0; i<l; i++)
        {
        fputc(LOW,f); 
        }
/* The "low" duration should be equal to the "high" one, so that
   it works even if the wires are inverted.
*/
for (i=0; i<l; i++)
        {
        fputc(HIGH,f); 
        }
}

/* Writes a byte to the .wav file */
void putbyte(FILE *f, unsigned char b, unsigned long int *filesize)
{
int i;


for (i=0; i<8; i++)
        {
        putbit(f, b & 128, filesize);
        b<<=1;
        }

}

void putblock(FILE *f, unsigned char *block, unsigned int size, unsigned long int *filesize)
{
int i;
for (i=0; i<size; i++)
        putbyte(f, block[i], filesize);
}

/* Creates the header for the .wav file */
void putWavHeader(FILE *f)
{

/* Header for .wav files. Dumbly copied from an existent .wav file :) */
/* (I haven't fully understood how it works :)))                      */

/* (03/09/97) Kev Thacker provided some help for these .wav files.
   Thanks to him. */



unsigned char WavHeader[44]=
{
0x52, 0x49, 0x46, 0x46, 0x72, 0xB8, 0x01, 0x00,
0x57, 0x41, 0x56, 0x45, 0x66, 0x6D, 0x74, 0x20,
0x10, 0x00, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00,
0x11, 0x2B, 0x00, 0x00, 0x11, 0x2B, 0x00, 0x00,
0x01, 0x00, 0x08, 0x00, 0x64, 0x61, 0x74, 0x61,
0x18, 0xB8, 0x01, 0x00
};

WavHeader[24]=rate&0xff;
WavHeader[25]=(rate >> 8) &0xff;
WavHeader[28]=rate&0xff;
WavHeader[29]=(rate >> 8) &0xff;


fwrite(WavHeader,1,44,f);
}


void writesize(FILE *f, unsigned long int size)
{
/* This might cause problems on little-endian systems. If
so, please let me know. */

unsigned long int s2;
fseek(f,40,SEEK_SET);
fwrite(&size,4,1,f); 
fseek(f,4,SEEK_SET);
s2=size + 36;        /* I think this is right. Not sure at all. :) */
fwrite(&s2,4,1,f); 
}

/* Here are the CRC functions ...
   Note: I DIDN'T try to understand them ! :)
*/
void putCRC(FILE *f, unsigned int CRC, unsigned long int *filesize)
{
unsigned char c1,c2;
c1 = (unsigned char)((CRC & 0xff00) >> 8);
c2 = (unsigned char)CRC & 0xff ;
putbyte(f, c1, filesize);
putbyte(f, c2, filesize);
}

unsigned int CRCupdate(unsigned int CRC, unsigned char new)
{
 unsigned int aux = CRC ^ (new << 8);
 int i;

 for(i=0; i<8; i++)
   if (aux & 0x8000)
       aux = (aux <<= 1) ^ kCRCpoly;
     else
       aux <<= 1;

 return(aux);
}


unsigned int calcCRC(unsigned char* buffer)
{

int j;
unsigned int CRC;


  CRC = 0xFFFF;
  for(j=0; j<256; j++)
    CRC = CRCupdate(CRC, buffer[j]);
  CRCupdate(CRC, 0);
  CRCupdate(CRC, 0);

 return((~CRC) & 0xffff);
}


void writename(unsigned char *tape_hdr, unsigned char *disc_hdr)
{
/* Processes the CPC file name.          */
/* Mainly, to manage file extensions ... */
int i=0,j=0;
while((disc_hdr[i+1] != 32)&&(i<8))
        {
        tape_hdr[i]=disc_hdr[i+1];
        i++;
        }
if (noext == 0)
        {
        if (disc_hdr[9]!=32) {tape_hdr[i]='.'; i++;} /* File extension present */
        while((disc_hdr[9+j] != 32) && (j<3))
                {
                tape_hdr[i]=disc_hdr[9+j];
                i++; j++;
                }
        }
}

void putsilence(FILE *f, int length, unsigned long int *filesize)
{
int i;
*filesize+=length;
for (i=0; i<length; i++) fputc(128,f);
}


void usage(char *progname)
{
        fprintf(stderr,"Usage: %s FILE [-evp][-r rate][-w prog][-o file][-s length]\n",progname);
        fprintf(stderr,"FILE : CPC file to be converted.");
        fprintf(stderr," -e : Don't create file extension.\n");
        fprintf(stderr," -v : Verbose mode.\n");
        fprintf(stderr," -p : Don't play file out.wav at end.\n");
        fprintf(stderr," -r <rate>: Specify output rate (Hz).\n");
        fprintf(stderr," -w <program>: Specify WAV player.\n");
        fprintf(stderr," -o <file>: Specify output file.\n");
        fprintf(stderr," -s <length>: Specify silence duration.\n");
        exit(1);
}



main(int argc, unsigned char **argv)

{


FILE *f,*source;
unsigned char header[256];
unsigned char srchead[128];
unsigned char srcdata[2048];
unsigned char data[256];
unsigned long int filesize=0;
unsigned int size;
unsigned int blocks;
unsigned int currentblock=1;
char filename[80];
char wavfile[80]="out.wav\0"; /* Default file name */
char player[80]="sndrec32\0"; /* Win95's sound player. */ 
char state='f';
char commande[100];
int i;
int j;
int nchunks;
filename[0]=0;
printf("CPC2TAPE version 1.1 1999 Pierre Thevenet\n");
silence=2*rate;



for (i=1;i<argc;i++)
{
switch(state)
{
case 'f':
        switch(argv[i][0])
        {
        case('-'):

                for (j=1;j < strlen(argv[i]); j++)
                {
                switch (argv[i][j])
                        {
                        case 'e': noext = 1; state='f'; break;
                        case 'v': verbose = 1; state='f'; break;
                        case 'p': play = 0; state='f'; break;
                        case 'r': state = 'r'; break;
                        case 'b': state = 'b'; break;
                        case 's': state = 's'; break;
                        case 'o': state = 'o'; break;
                        case 'w': state = 'w'; break;
                        default: fprintf(stderr,"Unrecognized option:%c. Ignored.\n",argv[2][i]); break;
                        }
                }
        break;
        default:
        if (filename[0]!=0)
                fprintf(stderr,"Only one file ! %s used (not %s)\n",filename,argv[i]);
        else
                strcpy(filename,argv[i]);
        break;
        }
        break;

case 'r':
        rate=atoi(argv[i]);
        state='f';
        if (rate==0)
        {
                fprintf(stderr,"Invalid rate: %s\n",argv[i]);
                fprintf(stderr,"Back to default rate (11025 Hz)\n");
                rate=11025;
        }
        printf("Using rate: %ud\n",rate);
        break;
case 'b':
        blank=atoi(argv[i]);
        state='f';
        if (blank==0)
        {
                fprintf(stderr,"Invalid blank value: %s\n",argv[i]);
                fprintf(stderr,"Back to default blank value (0)\n");
                blank=0;
        }
        printf("Using blank period: %ud\n",blank);
        break;
case 's':
        silence=atoi(argv[i]);
        state='f';
        if (silence==0)
        {
                fprintf(stderr,"Invalid silence: %s\n",argv[i]);
                fprintf(stderr,"Back to default silence (2 seconds)\n");
                silence=2*rate;
        }
        printf("Using silence: %ud\n",silence);
        break;
case 'o':
        strcpy(wavfile,argv[i]);
        state='f';
        printf("Using wav file: %s\n",argv[i]);
        break;
case 'w':
        strcpy(player,argv[i]);
        state='f';
        printf("Using player: %s\n",argv[i]);
        break;
default:
        fprintf(stderr,"What are you doing here ? This code should never be reached.\n");
        exit(6128);
        break;
} /* END-SWITCH */

} /* END-FOR */ /* Abuse of COBOL can seriously damage your health */


if (state=='r')
{
        fprintf(stderr,"Option -r should be followed by a rate.\n");
}

if (state=='s')
{
        fprintf(stderr,"Option -s should be followed by a value.\n");
}

if (state=='b')
{
        fprintf(stderr,"Option -b should be followed by a value.\n");
}

if (filename[0]==0)
{
        fprintf(stderr,"Cannot run without a file to convert !\n");
        usage(argv[0]);
}


if (!(source=fopen(filename,"rb")))
{
        fprintf(stderr,"Error opening file %s . Giving up.\n",filename);
        fprintf(stderr,"File probably doesn't exist\n");
        exit(2);
}


fread(srchead,128,1,source);
size=srchead[64]+srchead[65]*256;
if (verbose) printf("CPC file size (%s):%d bytes\n",filename,size);

blocks=size>>11;


if (size&2047) blocks++;
if (verbose) printf("Need %d blocks.\n",blocks);

if (!(f=fopen("out.wav", "wb")))
{
        /* This is not very likely to happen anyway ! */
        fprintf(stderr,"Fatal error: unable to open output file out.wav !\n");
        exit(2);
}
putWavHeader(f);

for (i=0; i<256; i++) header[i]=0;
header[24]=srchead[64];
header[25]=srchead[65];
header[17]=0;
writename(header,srchead);
for (i=18; i<28; i++) header[i]=srchead[i];
if (verbose) printf("Total size : %4x\n",header[24]+header[25]*256);
if (verbose) printf("Entry point: %4x\n",header[26]+header[27]*256);
if (verbose) printf("Origin     : %4x\n",header[21]+header[22]*256);

for (currentblock=1; currentblock <= blocks; currentblock++)
{
if (verbose) printf("Processing file %s block %d ",header,currentblock);

/* Filling in the tape header ... */
        header[16]=(unsigned char) currentblock;
        header[19] = 0; /* Default block size: 2048 bytes */
        header[20] = 8;


        if (currentblock == blocks)  /* Last block */
                {
                header[17] = 0xff;
                if (size % 2048)
                        {
                        header[20] = (size & 0x0700) >> 8;
                        header[19] = size & 0xff ;
                        }
                }
        if (currentblock == 1)  header[23] = 0xff;  /* First Block */
        /* AMSDOS won't load the file if this is not set to 255 !
        it will reply "Found FILE block n" , whatever the block number */
if (verbose) printf("Size of block:%d\n",header[19]+header[20]*256);
fread(srcdata,header[19]+header[20]*256,1,source);

if (feof(source))
{
	fprintf(stderr,"Fatal error: EOF met on input file.\nMaybe a non-CPC or ASCII file ?\n");
	exit(2);
}


/* Actual header writing */
        for (i=0; i<256; i++)  putbyte(f, 0xff, &filesize); /* leader */
        putbit(f,0,&filesize);      /* Sync bit & byte. For more details see */
        putbyte(f,0x2c,&filesize);  /* AIFFdec's refrnce.txt !               */
        putblock(f,header,256,&filesize);
        putCRC(f,calcCRC(header),&filesize);
        for (i=0; i<4; i++) putbyte(f,255,&filesize); /* trailer */
        putsilence(f,blank,&filesize);
        for (i=0; i<256; i++) putbyte(f,255,&filesize); /* leader */
        putbit(f,0,&filesize);
        putbyte(f,0x16,&filesize);

/* Block writing */

        nchunks=header[20];
        if (header[19]!=0) nchunks++;

        /* Number of chunks is block size/256, +1 if there is a remainder */
        for(j=0; j<nchunks; j++) 
        {
                putblock(f,srcdata+(j*256),256,&filesize);
                putCRC(f,calcCRC(srcdata+(j*256)),&filesize);
        }
        for (i=0; i<4; i++) putbyte(f,255,&filesize); /* trailer */
        if (currentblock!=blocks)
                putsilence(f,silence,&filesize);
}

if (verbose) printf("Output file size:%ld\n",filesize);
printf("Actual rate:%ld bits /sec\n",(long int)(((long int)(size*8*rate))/filesize));
writesize(f, filesize);
fclose(f); 
/* Calls the sound player. */
 sprintf(commande,"%s %s",player,wavfile);
 if (play) system(commande);


return(0);
}
